/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.communication.network;

import java.io.IOException;
import java.io.Serializable;
import java.net.URL;
import java.util.Hashtable;
import java.util.Map;

import javax.naming.ConfigurationException;

import org.jgroups.Address;
import org.jgroups.Channel;
import org.jgroups.ChannelClosedException;
import org.jgroups.ChannelFactory;
import org.jgroups.ChannelNotConnectedException;
import org.jgroups.Global;
import org.jgroups.JChannelFactory;
import org.jgroups.Message;
import org.jgroups.Receiver;
import org.jgroups.View;
import org.jgroups.blocks.ReplicatedHashMap;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.communication.network.data.DeleteReplicatedHashMapData;
import org.ow2.petals.communication.network.data.JoinRequestData;
import org.ow2.petals.communication.network.data.JoinResponseData;
import org.ow2.petals.communication.network.data.NetworkData;
import org.ow2.petals.communication.topology.TopologyService;
import org.ow2.petals.kernel.api.server.PetalsException;
import org.ow2.petals.kernel.configuration.ConfigurationService;
import org.ow2.petals.kernel.configuration.ContainerConfiguration;
import org.ow2.petals.kernel.configuration.DomainConfiguration;
import org.ow2.petals.kernel.configuration.SubDomainConfiguration;
import org.ow2.petals.topology.TopologyException;
import org.ow2.petals.util.LoggingUtil;

import static org.ow2.petals.kernel.configuration.DomainConfiguration.DomainMode.STANDALONE;

/**
 * This {@link NetworkService} implementation is based on JGROUPS.<br>
 * It uses 2 different stacks. One is based on UPD over multicast IP, the other
 * is based on TCP over unicast IP.<br>
 * The first stack is more efficient. For Networks whom multicast messages are
 * forbidden, the second stack must be used.
 * 
 * @author Roland NAUDIN - eBMWebSourcing
 * @since Petals 2.0
 * 
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.ow2.petals.communication.network.NetworkService.class))
public class NetworkServiceImpl implements NetworkService, Receiver {

    private static final String STACKS_CONFIG_FILE = "stacks.xml";

    // private static final String PERSIST_CONFIG_FILE = "persist.properties";

    private static final String UDP_MCAST_STACK = "udp";

    private static final String TCP_UCAST_STACK = "tcp";

    /**
     * The latency max value. If no response is received over this value, the
     * responder is supposed absent
     */
    private static final int LATENTY_MAX_TIME = 5000;

    /**
     * The error message when the latency max value has been reached.
     */
    private static final String TIMEOUT_REACHED_ERROR = "The coordinator of the group do not respond";

    /**
     * Logging utilities
     */
    private LoggingUtil log;

    @Monolog(name = "logger")
    private Logger logger;

    /**
     * The topology service fractal component
     */
    @Requires(name = "topology", signature = org.ow2.petals.communication.topology.TopologyService.class)
    private TopologyService topologyService;

    /**
     * The Configuration service fractal component
     */
    @Requires(name = "configuration", signature = org.ow2.petals.kernel.configuration.ConfigurationService.class)
    private ConfigurationService configurationService;

    /**
     * The container configuration
     */
    private ContainerConfiguration containerConfiguration;

    /**
     * The sub-domain configuration
     */
    private SubDomainConfiguration subDomainConfiguration;

    /**
     * The domain configuration
     */
    private DomainConfiguration domainConfiguration;

    /**
     * The JGROUPS channel factory
     */
    private ChannelFactory channelFactory;

    /**
     * A channel for the Discovery service
     */
    private Channel discoveryCoordinationChannel;

    /**
     * The replicated HashMaps used by the JNDI service to store its contexts
     * data.
     */
    private Map<String, ReplicatedHashMap<String, Serializable>> replicatedHashMaps;

    /**
     * The replicated HashMap container Addresses map
     */
    private ReplicatedHashMap<String, Address> containerAddressesHashMap;

    private String stackName;

    /**
     * The result of the joining request. Used to synchronized the response.
     */
    private String[] joinResult = new String[1];

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.communication.network.NetworkService#joinNetwork()
     */
    public String joinNetwork() throws NetworkException {
        this.log.start();

        synchronized (this.joinResult) {
            // always true if the node is the coordinator and so the first node
            if (this.discoveryCoordinationChannel.getView().getCreator().equals(
                    this.discoveryCoordinationChannel.getLocalAddress())) {
                this.joinResult[0] = JOIN_ACCEPTED;
            } else {
                this.joinResult[0] = TIMEOUT_REACHED_ERROR;
                NetworkData joinData = new JoinRequestData(this.containerConfiguration,
                        this.subDomainConfiguration, this.domainConfiguration);
                try {
                    this.discoveryCoordinationChannel.send(this.discoveryCoordinationChannel
                            .getView().getCreator(), null, joinData);
                    this.joinResult.wait(LATENTY_MAX_TIME);
                } catch (InterruptedException e) {
                    // do nothing
                } catch (ChannelNotConnectedException e) {
                    throw new NetworkException(e);
                } catch (ChannelClosedException e) {
                    throw new NetworkException(e);
                }
            }
        }

        this.log.end();
        return this.joinResult[0];
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.communication.network.NetworkService#getReplicatedHashMap
     * (java.lang.String)
     */
    public Map<String, Serializable> getReplicatedHashMap(String key) throws NetworkException {
        this.log.start(key);

        ReplicatedHashMap<String, Serializable> replicatedHashMap = this.replicatedHashMaps
                .get(key);

        if (replicatedHashMap == null) {
            try {
                Channel replicatedHashMapChannel = this.channelFactory.createMultiplexerChannel(
                        stackName, key);

                replicatedHashMap = new ReplicatedHashMap<String, Serializable>(
                        replicatedHashMapChannel);
                // the updates on the replicated hashMap must be synchronous
                replicatedHashMap.setBlockingUpdates(true);
                replicatedHashMap.setTimeout(LATENTY_MAX_TIME);

                // TODO: uncomment this part when the persistence is finalized
                // URL persistURL = this.getClass().getResource("/" +
                // PERSIST_CONFIG_FILE);
                // if (persistURL == null) {
                // log.warning(
                // "Failed to reach the JGroups persistence resource: '"
                // + PERSIST_CONFIG_FILE +
                // "'. The persistence of JNDI is turned off");
                // }
                // this.jndiHashMap.setPersistPropertiesFile(persistURL.getFile(
                // ));

                // start the JNDI HashMap and wait the synchronization of its
                // data
                replicatedHashMapChannel.connect(this.domainConfiguration.getName());
                replicatedHashMap.start(LATENTY_MAX_TIME);
                this.replicatedHashMaps.put(key, replicatedHashMap);
            } catch (Exception e) {
                throw new NetworkException("Failed to create Replicated Map for key '" + key + "'",
                        e);
            }
        }

        this.log.end();
        return replicatedHashMap;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.communication.network.NetworkService#deleteReplicatedHashMap
     * (java.lang.String)
     */
    public void deleteReplicatedHashMap(String key) throws NetworkException {
        this.log.start(key);

        ReplicatedHashMap<String, Serializable> replicatedHashMap = this.replicatedHashMaps
                .remove(key);

        try {
            if (replicatedHashMap == null) {
                this.discoveryCoordinationChannel.send(null, null, new DeleteReplicatedHashMapData(
                        key));
            } else {
                // synchronized to avoid local modification of the
                // ReplicatedHashMap during the deletion
                synchronized (replicatedHashMap) {
                    this.discoveryCoordinationChannel.send(null, null,
                            new DeleteReplicatedHashMapData(key));

                    replicatedHashMap.stop();
                }
            }

        } catch (ChannelNotConnectedException e) {
            throw new NetworkException(e);
        } catch (ChannelClosedException e) {
            throw new NetworkException(e);
        }

        this.log.end();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.jgroups.MessageListener#receive(org.jgroups.Message)
     */
    public void receive(Message message) {
        this.log.start();

        // if the message has a destination, it means that it is a join request/
        // response
        if (message.getObject() instanceof NetworkData) {
            NetworkData networkData = (NetworkData) message.getObject();

            if (this.discoveryCoordinationChannel.getLocalAddress().equals(message.getDest())) {
                if (networkData.getType() == NetworkData.MessageType.joinRequest) {
                    this.log
                            .info("A PEtALS container tries to join the Network, check its configuration");

                    ContainerConfiguration newContainerConfiguration = ((JoinRequestData) message
                            .getObject()).getContainerConfiguration();
                    SubDomainConfiguration newSubDomainConfiguration = ((JoinRequestData) message
                            .getObject()).getSubDomainConfiguration();
                    DomainConfiguration newDomainConfiguration = ((JoinRequestData) message
                            .getObject()).getDomainConfiguration();
                    String response = JOIN_ACCEPTED;
                    try {
                        this.configurationService.verifyContainerTopology(
                                newContainerConfiguration, newSubDomainConfiguration,
                                newDomainConfiguration);
                    } catch (TopologyException e) {
                        this.log.warning("Refuse a PEtALS container named '"
                                + newContainerConfiguration.getName() + "' on "
                                + newContainerConfiguration.getHost() + " : " + e.getMessage());
                        response = e.getMessage();
                    }

                    if (JOIN_ACCEPTED.equals(response)) {
                        // refuse the candidate if the container is already
                        // registered as started
                        if (this.containerAddressesHashMap.containsKey(newContainerConfiguration
                                .getName())) {
                            response = "Container named '" + newContainerConfiguration.getName()
                                    + "' is already started";
                            this.log.warning("Refuse a PEtALS container named '"
                                    + newContainerConfiguration.getName() + "' on "
                                    + newContainerConfiguration.getHost() + " : " + response);

                        } else {
                            // update the container addresses map
                            this.containerAddressesHashMap.put(newContainerConfiguration.getName(),
                                    message.getSrc());
                            this.log.info("Accept the PEtALS container named '"
                                    + newContainerConfiguration.getName() + "' on "
                                    + newContainerConfiguration.getHost());
                        }
                    }

                    NetworkData responseData = new JoinResponseData(response);
                    try {
                        this.discoveryCoordinationChannel
                                .send(message.getSrc(), null, responseData);
                    } catch (ChannelNotConnectedException e) {
                        this.log.error("Coordinator failed to send the join response",
                                new NetworkException(e));
                    } catch (ChannelClosedException e) {
                        this.log.error("Coordinator failed to send the join response",
                                new NetworkException(e));
                    }
                } else if (networkData.getType() == NetworkData.MessageType.joinResponse) {
                    synchronized (this.joinResult) {
                        this.joinResult[0] = ((JoinResponseData) message.getObject()).getResponse();
                        this.joinResult.notify();
                    }
                }
            } else if (networkData.getType() == NetworkData.MessageType.deleteReplicatedHashMap) {
                // Delete the ReplicatedHasMap
                String key = ((DeleteReplicatedHashMapData) networkData).getKey();
                ReplicatedHashMap<String, Serializable> replicatedHashMap = this.replicatedHashMaps
                        .remove(key);
                if (replicatedHashMap != null) {
                    replicatedHashMap.stop();
                }
            }
        }

        this.log.end();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.jgroups.MembershipListener#viewAccepted(org.jgroups.View)
     */
    public void viewAccepted(View new_view) {
        this.log.start();

        // test if this node is the coordinator
        if (new_view.getMembers().get(0)
                .equals(this.discoveryCoordinationChannel.getLocalAddress())) {
            if (this.containerAddressesHashMap.size() == 0) {
                // add first element of the group in the map
                this.containerAddressesHashMap.put(this.containerConfiguration.getName(),
                        this.discoveryCoordinationChannel.getLocalAddress());
            } else if (this.containerAddressesHashMap.size() > new_view.getMembers().size()) {
                // remove the dead nodes and update their state in the topology
                for (String containerName : this.containerAddressesHashMap.keySet()) {
                    if (!new_view.getMembers().contains(
                            this.containerAddressesHashMap.get(containerName))) {
                        // update the topology
                        try {
                            this.topologyService.setContainerState(containerName,
                                    ContainerConfiguration.ContainerState.STOPPED);
                            this.log
                                    .info("Stop the PEtALS container named '" + containerName + "'");
                        } catch (PetalsException e) {
                            // in dynamic mode, the container may have leaved
                            // the domain (shut down)
                            if (!domainConfiguration.getMode().equals(
                                    DomainConfiguration.DomainMode.DYNAMIC)) {
                                this.log.error(
                                        "Failed to update the state to STOPPED of the container named '"
                                                + containerName + "'", e);
                            }
                        }
                        this.containerAddressesHashMap.remove(containerName);
                    }
                }
            }
        }

        this.log.end();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.jgroups.MessageListener#getState()
     */
    public byte[] getState() {
        // do nothing
        return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.jgroups.MessageListener#setState(byte[])
     */
    public void setState(byte[] arg0) {
        // do nothing

    }

    /*
     * (non-Javadoc)
     * 
     * @see org.jgroups.MembershipListener#block()
     */
    public void block() {
        // do nothing

    }

    /*
     * (non-Javadoc)
     * 
     * @see org.jgroups.MembershipListener#suspect(org.jgroups.Address)
     */
    public void suspect(Address arg0) {
        // do nothing

    }

    @LifeCycle(on = LifeCycleType.START)
    protected void start() throws Exception {
        this.log = new LoggingUtil(this.logger);
        this.log.start();

        this.containerConfiguration = this.configurationService.getContainerConfiguration();
        this.subDomainConfiguration = this.configurationService.getSubDomainConfiguration();
        this.domainConfiguration = this.configurationService.getDomainConfiguration();
        this.replicatedHashMaps = new Hashtable<String, ReplicatedHashMap<String, Serializable>>();

        try {
            this.createChannels();

            this.startChannels();
        } catch (Exception e) {
            this.log.error("Failed to start Network service", e);
            throw e;
        }

        this.log.end();
    }

    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() {
        this.log.start();

        if (!STANDALONE.equals(this.domainConfiguration.getMode())) {

            for (ReplicatedHashMap<String, Serializable> replicatedHashMap : this.replicatedHashMaps
                    .values()) {
                Channel replicatedHashMapChannel = replicatedHashMap.getChannel();
                if (replicatedHashMapChannel.isConnected()) {
                    replicatedHashMap.stop();
                }
            }

            Channel containerAddressesHashMapChannel = this.containerAddressesHashMap.getChannel();
            if (containerAddressesHashMapChannel.isConnected()) {
                this.containerAddressesHashMap.stop();
            }

            if (this.discoveryCoordinationChannel.isConnected()) {
                this.discoveryCoordinationChannel.close();
            }
        }

        this.log.end();
    }

    /**
     * Create the JGROUPS channels
     * 
     * @throws Exception
     */
    private void createChannels() throws Exception {
        this.log.start();

        initChannels();

        this.channelFactory = new JChannelFactory();
        URL stacksURL = this.getClass().getResource("/" + STACKS_CONFIG_FILE);
        if (stacksURL == null) {
            throw new IOException("Failed to reach the JGroups ressource '" + STACKS_CONFIG_FILE
                    + "'");
        }

        this.channelFactory.setMultiplexerConfig(stacksURL);

        this.discoveryCoordinationChannel = this.channelFactory.createMultiplexerChannel(stackName,
                "A");
        this.discoveryCoordinationChannel.setReceiver(this);

        Channel discoveryHashMapChannel = this.channelFactory.createMultiplexerChannel(stackName,
                "B");
        this.containerAddressesHashMap = new ReplicatedHashMap<String, Address>(
                discoveryHashMapChannel);

        this.log.end();
    }

    /**
     * Initialize the JGROUPS channel and connect them to the group.
     * 
     * @throws Exception
     */
    private void startChannels() throws Exception {
        this.log.start();

        // start the discovery HashMap and wait the synchronization of its data
        this.containerAddressesHashMap.getChannel().connect(this.domainConfiguration.getName());
        this.containerAddressesHashMap.start(LATENTY_MAX_TIME);

        this.discoveryCoordinationChannel.connect(this.domainConfiguration.getName());

        this.log.end();
    }

    /**
     * Initialize the channels environments and return the stack configuration.
     * 
     * @return
     * @throws IOException
     * @throws ConfigurationException
     * @throws NoSuchFieldException
     * @throws SecurityException
     * @throws IllegalAccessException
     * @throws IllegalArgumentException
     */
    private void initChannels() throws IOException, ConfigurationException, SecurityException,
            NoSuchFieldException, IllegalArgumentException, IllegalAccessException {

        // Global configuration
        System.setProperty(Global.BIND_ADDR, String.valueOf(this.containerConfiguration.getHost()));

        System.setProperty("jgroups.bind_port", String.valueOf(this.containerConfiguration
                .getNetworkBindingPort()));

        // TODO: Finalize the JNDI persistence to uncomment this part
        // File workDirectory = new File(containerConfiguration.
        // getInstallDirectoryPath()
        // + File.separator + "work");
        // File persistFile = new File(workDirectory, "persist.properties");
        // if (!persistFile.exists()) {
        // String jndiPersistenceDir = new String(workDirectory.getAbsolutePath(
        // ) + File.separator
        // + "jndi");
        // FileOutputStream output = new FileOutputStream(persistFile);
        // output.write("persist=".getBytes());
        // output.write(jndiPersistenceDir.getBytes());
        // output.write("\nfilePersistMgr=".getBytes());
        // output.write(FilePersistenceManager.class.getName().getBytes());
        // output.close();
        // }

        // Configure either the TCP_UCAST_STACK or the UDP_MCAST_STACK

        if (this.domainConfiguration.getNetworkUnicastAddressMap() != null) {
            stackName = TCP_UCAST_STACK;
            String networkUnicastAdressList = "";
            Map<String, String> networkUnicastAddressMap = this.domainConfiguration
                    .getNetworkUnicastAddressMap();
            for (String networkUnicastAddress : networkUnicastAddressMap.values()) {
                networkUnicastAdressList = networkUnicastAdressList + networkUnicastAddress + ",";
            }
            networkUnicastAdressList.substring(0, networkUnicastAdressList.length() - 1);
            System.setProperty(Global.TCPPING_INITIAL_HOSTS, networkUnicastAdressList);
        } else if (this.domainConfiguration.getNetworkMuticastIp() != null) {
            stackName = UDP_MCAST_STACK;
            System.setProperty(Global.UDP_MCAST_ADDR, this.domainConfiguration
                    .getNetworkMuticastIp());
            System.setProperty(Global.UDP_MCAST_PORT, String.valueOf(this.domainConfiguration
                    .getNetworkMuticastPort()));
        } else {
            throw new ConfigurationException("The network configuration is missing");
        }
    }
}
