/**
 * PETALS: PETALS Services Platform Copyright (C) 2005 EBM WebSourcing
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 * 
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: LifeCycle.java,v 1.2 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------
 */

package org.ow2.petals.container.lifecycle;

import javax.jbi.JBIException;
import javax.jbi.management.LifeCycleMBean;
import javax.management.AttributeChangeNotification;
import javax.management.ListenerNotFoundException;
import javax.management.MBeanNotificationInfo;
import javax.management.Notification;
import javax.management.NotificationBroadcaster;
import javax.management.NotificationBroadcasterSupport;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.ObjectName;

import org.ow2.petals.container.event.StateChangeFailedEvent;
import org.ow2.petals.container.event.StateChangedEvent;
import org.ow2.petals.util.LoggingUtil;
import org.ow2.petals.util.PetalsRuntimeException;

/**
 * Implementation of the <code>LifeCycle</code> mbean. This lifecycle mbean
 * sends notifications about changes of state.
 * 
 * @author Adrien LOUIS - EBM WebSourcing
 */
public abstract class LifeCycleAbstract implements LifeCycleMBean, NotificationBroadcaster {

    protected Object activitySynchronizer;

    protected boolean initialized;

    protected LoggingUtil log;

    protected ObjectName mBeanName;

    protected long sequenceNumber;

    protected String state;

    protected String previousState;

    protected String name;

    NotificationBroadcasterSupport notifSupport = new NotificationBroadcasterSupport();

    /**
     * Creates a new instance of {@link LifeCycleAbstract}
     * 
     */
    public LifeCycleAbstract() {
        state = UNKNOWN;
    }

    /**
     * 
     * @param mBeanName
     * @param log
     */
    public void init(ObjectName mBeanName, LoggingUtil log) {
        this.log = log;
        log.call();
        // new Byte[0] is supposed to be the lighter instance to create
        // according to some specialized documentation
        activitySynchronizer = new Byte[0];
        state = SHUTDOWN;
        initialized = false;
        this.mBeanName = mBeanName;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * javax.management.NotificationBroadcaster#addNotificationListener(javax
     * .management.NotificationListener, javax.management.NotificationFilter,
     * java.lang.Object)
     */
    public void addNotificationListener(NotificationListener arg0, NotificationFilter arg1,
            Object arg2) throws IllegalArgumentException {
        notifSupport.addNotificationListener(arg0, arg1, arg2);
    }

    /**
     * 
     * @throws JBIException
     */
    public abstract void doInit() throws JBIException;

    /**
     * 
     * @throws JBIException
     */
    public abstract void doShutdown() throws JBIException;

    /**
     * 
     * @throws JBIException
     */
    public abstract void doStart() throws JBIException;

    /**
     * 
     * @throws JBIException
     */
    public abstract void doStop() throws JBIException;

    protected abstract String getJBIArtifactType();

    /**
     * @see javax.jbi.management.LifeCycleMBean#getCurrentState()
     */
    public String getCurrentState() {
        synchronized (this.state) {
            if (!(SHUTDOWN.equals(state) || STARTED.equals(state) || STOPPED.equals(state) || UNKNOWN
                    .equals(state))) {
                return UNKNOWN;
            }
        }
        return state;
    }

    /**
     * Get the MBean name
     * 
     * @return
     */
    public ObjectName getMBeanName() {
        return mBeanName;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.management.NotificationBroadcaster#getNotificationInfo()
     */
    public MBeanNotificationInfo[] getNotificationInfo() {
        String[] types = new String[] { AttributeChangeNotification.ATTRIBUTE_CHANGE };

        String name = AttributeChangeNotification.class.getName();

        String description = "Notification about a change of the lifeCycle state or a state change failure.";

        MBeanNotificationInfo info = new MBeanNotificationInfo(types, name, description);

        return new MBeanNotificationInfo[] { info };
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * javax.management.NotificationBroadcaster#removeNotificationListener(javax
     * .management.NotificationListener)
     */
    public void removeNotificationListener(NotificationListener arg0)
            throws ListenerNotFoundException {
        notifSupport.removeNotificationListener(arg0);
    }

    public void setMBeanName(ObjectName beanName) {
        mBeanName = beanName;
    }

    /**
     * Set the <code>State</code> reference of the current state. An
     * <code>LifeCycleStateEventAbstract</code> is thrown.
     * 
     * @param state
     */
    public void setState(String state) throws JBIException {
        synchronized (this.state) {
            if (!(SHUTDOWN.equals(state) || STARTED.equals(state) || STOPPED.equals(state) || UNKNOWN
                    .equals(state))) {
                throw new PetalsRuntimeException("State '" + state
                        + "' isn't defined by the JBI specification. No state changement done.");
            }
            this.previousState = new String(this.state);
            stateChanged(new StateChangedEvent(this, this.state, state));
            this.state = state;
        }
    }

    /**
     * Shutdown the component. A notification will be thrown in case of
     * runtimeException or incoherent state.
     * 
     * @see javax.jbi.management.LifeCycleMBean#shutdown()
     * @throws JBIException
     *             in case of runtimeException or incoherent state.
     */
    public void shutDown() throws JBIException {
        log.start();
        if (SHUTDOWN.equals(state)) {
            log.info("The component '" + this.name
                    + "' is already shutdown. No state changement done.");
        } else if (STOPPED.equals(state)) {
            synchronized (activitySynchronizer) {
                try {
                    doShutdown();
                    setState(SHUTDOWN);
                    initialized = false;
                } catch (Throwable re) {
                    setState(UNKNOWN);
                    JBIException e = new JBIException(
                            "An exception occured while shutting down the " + getJBIArtifactType()
                                    + " '" + this.name + "'", re);
                    log.error(e.getMessage(), re);
                    stateChangeFailed(new StateChangeFailedEvent(this, e));
                    throw e;
                }
            }
        } else if (STARTED.equals(state)) {
            this.stop();
            this.shutDown();
        } else {
            JBIException e = new JBIException("The " + getJBIArtifactType() + " '" + this.name
                    + "' can not be shut down in this state: " + state);
            log.error(e);
            stateChangeFailed(new StateChangeFailedEvent(this, e));
            throw e;
        }
        log.end();
    }

    /**
     * Start the component. A notification will be thrown in case of
     * runtimeException or incoherent state.
     * 
     * @see javax.jbi.management.LifeCycleMBean#start()
     * @throws JBIException
     *             in case of runtimeException or incoherent state.
     */
    public void start() throws JBIException {
        log.start();
        if (STARTED.equals(state)) {
            log.info("The component '" + this.name
                    + "' is already started. No state changement done.");
        } else if (SHUTDOWN.equals(state) || STOPPED.equals(state)) {
            synchronized (activitySynchronizer) {
                try {
                    if (!initialized) {
                        doInit();
                        initialized = true;
                    }
                    doStart();
                    setState(STARTED);
                } catch (Throwable re) {
                    setState(UNKNOWN);
                    JBIException e = new JBIException("An exception occured while starting the "
                            + getJBIArtifactType() + " '" + this.name + "'", re);
                    log.error(e.getMessage(), re);
                    stateChangeFailed(new StateChangeFailedEvent(this, e));
                    throw e;
                }
            }
        } else {
            JBIException e = new JBIException("The " + getJBIArtifactType() + " '" + this.name
                    + "' can not be started in this state: " + state);
            log.error(e);
            stateChangeFailed(new StateChangeFailedEvent(this, e));
            throw e;
        }

        log.end();
    }

    /**
     * <i>stateChanged</i> is called when the state of a lifecycle element
     * changes. A JMX-notification is thrown.
     * 
     * @see petals.management.element.ElementStateListener#stateChanged(java.util.EventObject)
     */
    protected void stateChanged(StateChangedEvent event) {
        log.call();
        if (mBeanName != null) {
            synchronized (this.state) {
                Notification n = new AttributeChangeNotification(mBeanName, sequenceNumber++,
                        System.currentTimeMillis(), "State of '" + mBeanName.toString()
                                + "' changed : " + event.getOldState() + " -> "
                                + event.getNewState(), "state", "String", event.getOldState(),
                        event.getNewState());
                notifSupport.sendNotification(n);
            }
        }
    }

    /**
     * <i>stateChangeFailed</i>> is called when the change of state of a
     * lifecycle element fails. A JMX-notification throws the relative
     * JBIException, .
     * 
     * @see petals.management.element.ElementStateListener#stateChanged(java.util.EventObject)
     */
    protected void stateChangeFailed(StateChangeFailedEvent event) {
        log.call();
        if (mBeanName != null) {
            synchronized (this.state) {
                Notification n = new AttributeChangeNotification(mBeanName, sequenceNumber++,
                        System.currentTimeMillis(), "State change of '" + mBeanName.toString()
                                + "' failed : " + event.getException().getMessage(), "exception",
                        "String", null, event.getException().toString());

                notifSupport.sendNotification(n);
            }
        }
    }

    /**
     * Stop the component. A notification will be thrown in case of
     * runtimeException or incoherent state.
     * 
     * @see javax.jbi.management.LifeCycleMBean#stop()
     * @throws JBIException
     *             in case of runtimeException or incoherent state.
     */
    public void stop() throws JBIException {
        log.start();
        if (STOPPED.equals(state)) {
            log.info("The component '" + this.name
                    + "' is already stopped. No state changement done.");
        } else if (STARTED.equals(state)) {
            synchronized (activitySynchronizer) {
                try {
                    doStop();
                    setState(STOPPED);
                } catch (UnauthorizedJBIException re) {
                    //A not allowed action was done. No state change.
                    JBIException e = new JBIException("The stop was not allowed by PEtALS", re);
                    log.info(e.getMessage(), re);
                    stateChangeFailed(new StateChangeFailedEvent(this, e));
                    throw e;
                }                
                catch (Throwable re) {
                    setState(UNKNOWN);
                    JBIException e = new JBIException("An exception occured while stopping the "
                            + getJBIArtifactType() + " '" + this.name + "'", re);
                    log.error(e.getMessage(), re);
                    stateChangeFailed(new StateChangeFailedEvent(this, e));
                    throw e;
                }
            }
        } else {
            JBIException e = new JBIException("The " + getJBIArtifactType() + " '" + this.name
                    + "' can not be stopped in this state: " + state);
            log.error(e);
            stateChangeFailed(new StateChangeFailedEvent(this, e));
            throw e;
        }
        log.end();
    }
}
