/**
 * PETALS - PETALS Services Platform. Copyright (c) 2005 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id: InstallationServiceMBean.java 154 2006-03-27 15:30:10Z ofabre $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.container.lifecycle;

import java.net.URI;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import javax.jbi.JBIException;
import javax.management.ObjectName;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.container.event.StateChangeFailedEvent;
import org.ow2.petals.jbi.descriptor.original.generated.ServiceAssembly;
import org.ow2.petals.kernel.api.server.PetalsException;
import org.ow2.petals.platform.systemstate.SystemStateService;
import org.ow2.petals.util.LoggingUtil;

/**
 * This class manage the life cycle of deployed service assemblies
 * 
 * @author Olivier Fabre - EBM Websourcing
 * @author ddesjardins - eBMWebsourcing
 */
@FractalComponent
public class ServiceAssemblyLifeCycle extends LifeCycleAbstract {

    private URI installationRoot;

    /*
     * Force the set of a state, ignoring eventual SUs life cycle errors
     */
    private boolean forcedState = false;

    @Monolog(name = "logger")
    protected Logger logger;

    /**
     * Platform component :: SystemState Service
     */
    @Requires(name = "systemstate", signature = org.ow2.petals.platform.systemstate.SystemStateService.class)
    private SystemStateService recoverySrv;

    /**
     * The component
     */
    @Service
    private Component comp;

    private ServiceAssembly serviceAssembly;

    private Map<String, ServiceUnitLifeCycle> serviceUnitsLifeCycles;

    public ServiceAssemblyLifeCycle() {
        super();
    }

    /**
     * Accessor to the fractal component
     * 
     * @return the fractal component
     */
    public Component getComponent() {
        return comp;
    }

    public void setForcedState(boolean forcedState) {
        this.forcedState = forcedState;
    }

    /**
     * 
     * @param mbeanName
     *            name of the MBean for this LifeCycle. In fact, the
     *            DeploymentServiceMBean name is used as there is no
     *            SALifeCycleMBean
     * @param log
     * @param serviceAssembly
     * @param repositoryService
     * @param systemStateService
     * @param installationRoot
     */
    public void init(ObjectName mbeanName, ServiceAssembly serviceAssembly, URI installationRoot) {
        super.init(mbeanName, log);

        this.installationRoot = installationRoot;
        this.serviceAssembly = serviceAssembly;
        this.name = serviceAssembly.getIdentification().getName();
        this.serviceUnitsLifeCycles = Collections
                .synchronizedMap(new LinkedHashMap<String, ServiceUnitLifeCycle>());
    }

    /**
     * Do nothing
     */
    @Override
    public void doInit() throws JBIException {
        // never called
    }

    @Override
    public void doShutdown() throws JBIException {
        Map<String, ServiceUnitLifeCycle> sus = this.getServiceUnitsLifeCycles();
        ServiceUnitLifeCycle suLifeCycle = null;
        List<ServiceUnitLifeCycle> suLifeCycles = null;

        try {
            synchronized (sus) {
                suLifeCycles = new ArrayList<ServiceUnitLifeCycle>(sus.values());
                Collections.reverse(suLifeCycles);
                for (Iterator<ServiceUnitLifeCycle> iter = suLifeCycles.iterator(); iter.hasNext();) {
                    suLifeCycle = iter.next();
                    try {
                        if (!SHUTDOWN.equals(suLifeCycle.getCurrentState())) {
                            suLifeCycle.shutDown();
                        }
                    } catch (JBIException e) {
                        if (!this.forcedState) {
                            throw e;
                        }
                        // continue, ignoring the error
                    }
                }
            }
        } catch (JBIException e) {
            this.log.warning("The Service Assembly '" + this.name
                    + "' failed to shut down, try to rollback its state");
            // try to get back to the previous state for each SUs already
            // processed
            this.log.warning("The Service Assembly");
            this.state = this.previousState;
            if (suLifeCycles != null) {
                Iterator<ServiceUnitLifeCycle> iter = suLifeCycles.iterator();
                ServiceUnitLifeCycle processedSuLifeCycle = iter.next();
                while (!processedSuLifeCycle.equals(suLifeCycle)) {
                    try {
                        suLifeCycle.start();
                        suLifeCycle.stop();
                    } catch (JBIException e1) {
                        this.state = UNKNOWN;
                    }
                    processedSuLifeCycle = iter.next();
                }
            }
            throw e;
        }
    }

    @Override
    public void doStart() throws JBIException {
        Map<String, ServiceUnitLifeCycle> sus = this.getServiceUnitsLifeCycles();
        ServiceUnitLifeCycle suLifeCycle = null;
        List<ServiceUnitLifeCycle> suLifeCycles = null;

        try {
            synchronized (sus) {
                suLifeCycles = new ArrayList<ServiceUnitLifeCycle>(sus.values());
                for (Iterator<ServiceUnitLifeCycle> iter = suLifeCycles.iterator(); iter.hasNext();) {
                    suLifeCycle = iter.next();
                    try {
                        if (!STARTED.equals(suLifeCycle.getCurrentState())) {
                            suLifeCycle.start();
                        }
                    } catch (JBIException e) {
                        if (!this.forcedState) {
                            throw e;
                        }
                        // continue, ignoring the error
                    }
                }
            }
        } catch (JBIException e) {
            // try to get back to the previous state for each SUs already
            // processed
            this.log.warning("The Service Assembly '" + this.name
                    + "' failed to start, try to rollback its state");
            this.state = this.previousState;
            if (suLifeCycles != null) {
                Iterator<ServiceUnitLifeCycle> iter = suLifeCycles.iterator();
                ServiceUnitLifeCycle processedSuLifeCycle = iter.next();
                while (!processedSuLifeCycle.equals(suLifeCycle)) {
                    try {
                        if (STOPPED.equals(this.previousState)) {
                            suLifeCycle.stop();
                        } else {
                            // reset to the initial state (at the installation
                            // of
                            // the SA)
                            suLifeCycle.shutDown();
                        }
                    } catch (JBIException e1) {
                        this.state = UNKNOWN;
                    }
                    processedSuLifeCycle = iter.next();
                }
            }
            throw e;
        }
    }

    @Override
    public void doStop() throws JBIException {
        Map<String, ServiceUnitLifeCycle> sus = this.getServiceUnitsLifeCycles();
        ServiceUnitLifeCycle suLifeCycle = null;
        List<ServiceUnitLifeCycle> suLifeCycles = null;

        try {
            synchronized (sus) {
                suLifeCycles = new ArrayList<ServiceUnitLifeCycle>(sus.values());
                Collections.reverse(suLifeCycles);
                for (Iterator<ServiceUnitLifeCycle> iter = suLifeCycles.iterator(); iter.hasNext();) {
                    suLifeCycle = iter.next();

                    try {
                        if (STARTED.equals(suLifeCycle.getCurrentState())) {
                            suLifeCycle.stop();
                        }
                    } catch (JBIException e) {
                        if (!this.forcedState) {
                            throw e;
                        }
                        // continue, ignoring the error
                    }
                }
            }
        } catch (JBIException e) {
            // try to get back to the previous state for each SUs already
            // processed
            this.log.warning("The Service Assembly '" + this.name
                    + "' failed to stop, try to rollback its state");
            this.state = this.previousState;
            if (suLifeCycles != null) {
                Iterator<ServiceUnitLifeCycle> iter = suLifeCycles.iterator();
                ServiceUnitLifeCycle processedSuLifeCycle = iter.next();
                while (!processedSuLifeCycle.equals(suLifeCycle)) {
                    try {
                        suLifeCycle.start();
                    } catch (JBIException e1) {
                        this.state = UNKNOWN;
                    }
                    processedSuLifeCycle = iter.next();
                }
            }
            throw e;
        }
    }

    public URI getInstallationRoot() {
        return installationRoot;
    }

    public ServiceAssembly getServiceAssembly() {
        return serviceAssembly;
    }

    /**
     * Return the synchronized map of deployed service units.
     * 
     * @return
     */
    public Map<String, ServiceUnitLifeCycle> getServiceUnitsLifeCycles() {
        return serviceUnitsLifeCycles;
    }

    // ////////////////////////////////////////////
    // ElementListener implementation
    // ////////////////////////////////////////////

    /**
     * Register a ServiceUnitLifeCycle into a synchronized map of the
     * ServiceAssemblyLifeCycle
     * 
     * @param suName
     *            name of a service unit
     * @param serviceUnitLifeCycle
     *            life cycle of the service unit
     * @throws JBIException
     */
    public void registerSUIntoSA(String suName, ServiceUnitLifeCycle serviceUnitLifeCycle)
            throws JBIException {
        Map<String, ServiceUnitLifeCycle> sus = getServiceUnitsLifeCycles();
        synchronized (sus) {
            if (!sus.containsKey(suName)) {
                sus.put(suName, serviceUnitLifeCycle);
            } else {
                throw new JBIException("Duplicate service unit : same name");
            }

        }
    }

    /**
     * Unregister a ServiceUnitLifeCycle from a synchronized map of the
     * ServiceAssemblyLifeCycle
     * 
     * @param suName
     *            name of a service unit
     */
    public void unregisterSUFromSA(String suName) {
        Map<String, ServiceUnitLifeCycle> sus = getServiceUnitsLifeCycles();
        synchronized (sus) {
            sus.remove(suName);
        }
    }

    @Override
    public void setState(String state) throws JBIException {
        super.setState(state);
        try {
            recoverySrv.updateServiceAssemblyState(this.name, state);
        } catch (Exception e) {
            throw new JBIException("Failed to persist the state of the Service Assembly '"
                    + this.name + "'", e);
        }
    }

    /**
     * No init state for ServiceAssemblyLifeCycle
     */
    @Override
    public void start() throws JBIException {
        log.start();

        if (SHUTDOWN.equals(state) || STOPPED.equals(state)) {
            synchronized (activitySynchronizer) {
                try {
                    doStart();
                    setState(STARTED);
                } catch (Throwable re) {
                    setState(UNKNOWN);
                    JBIException e = new JBIException(
                            "An exception occured while shutting down the " + getJBIArtifactType()
                                    + " '" + this.name + "'", re);
                    log.error(e);
                    stateChangeFailed(new StateChangeFailedEvent(this, e));
                    throw e;
                }
            }
        } else {
            JBIException e = new JBIException("The " + getJBIArtifactType() + " '" + this.name
                    + "' can not be shutdown in this state: " + state);
            log.error(e);
            stateChangeFailed(new StateChangeFailedEvent(this, e));
            throw e;
        }
        log.end();
    }

    /**
     * 
     * @throws PetalsException
     * @throws JBIException
     * 
     */
    @LifeCycle(on = LifeCycleType.START)
    public void startComponentLifeCycle() {
        log = new LoggingUtil(logger, logger.getName().substring(
                logger.getName().lastIndexOf(".") + 1));
        log.call();
    }

    /**
     * Stop of the petals component
     * 
     */
    @LifeCycle(on = LifeCycleType.STOP)
    public void stopComponentLifeCycle() throws JBIException {
        log.call();
        // there is ambiguity between the STOP of the life-cycle and the STOP of
        // the fractal component, so stop the component only if the life-cycle
        // state is STARTED
        if (STARTED.equals(this.getCurrentState())) {
            this.doStop();
        }
    }

    @Override
    protected String getJBIArtifactType() {
        return "Service Assembly";
    }
}
