/**
 * PETALS - PETALS Services Platform. Copyright (c) 2006 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id: UndeployAllSUsTask.java 154 10 oct. 06 ofabre $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.management.task.deployment.undeploy;

import java.io.InputStream;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;

import javax.jbi.JBIException;
import javax.jbi.component.ServiceUnitManager;
import javax.jbi.management.DeploymentException;
import javax.jbi.management.LifeCycleMBean;

import org.ow2.petals.container.lifecycle.ComponentLifeCycle;
import org.ow2.petals.container.lifecycle.ServiceAssemblyLifeCycle;
import org.ow2.petals.container.lifecycle.ServiceUnitLifeCycle;
import org.ow2.petals.jbi.descriptor.original.generated.ServiceUnit;
import org.ow2.petals.jbi.management.admin.AdminServiceMBean;
import org.ow2.petals.jbi.management.task.AbstractLoggableTask;
import org.ow2.petals.jbi.management.task.deployment.DeploymentContextConstants;
import org.ow2.petals.jbi.management.task.deployment.DeploymentUtils;
import org.ow2.petals.jbi.management.util.XMLResult;
import org.ow2.petals.util.LoggingUtil;

/**
 * This task executes the undeployment process of a service unit, for each
 * Service Unit included into the service assembly <br/>
 * Inputs : {@link ServiceAssemblyLifeCycle}, {@link XMLResult}<br />
 * The task has got a mixed approach: it tries to do all the subtasks<br />
 * and doesn't have the undo capability. But if the service assembly isn't over
 * shutdown or unknown state,<br />
 * to implement the JBI specification, it stops and thrown an exception to
 * launch the undo process
 * 
 * @author ofabre - EBM Websourcing
 * @author Frederic Gardes - eBM Websourcing
 */
public class UndeployAllSUsTask extends AbstractLoggableTask {

    /**
     * JBI Container :: JMX Agent Admin
     */
    protected AdminServiceMBean managerService;

    public UndeployAllSUsTask(LoggingUtil log, AdminServiceMBean managerService) {
        super(log);
        this.managerService = managerService;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.processor.Task#execute(java.util.HashMap)
     */
    public void execute(HashMap<String, Object> context) throws Exception {
        if (context.containsKey(DeploymentContextConstants.SA_LIFECYCLE)) {
            ServiceAssemblyLifeCycle saLifeCycle = (ServiceAssemblyLifeCycle) context
                    .get(DeploymentContextConstants.SA_LIFECYCLE);
            if (saLifeCycle != null && context.containsKey(DeploymentContextConstants.XML_RESULT)) {
                XMLResult xmlResult = (XMLResult) context
                        .get(DeploymentContextConstants.XML_RESULT);
                if (xmlResult != null) {
                    // Undeploy SUs
                    undeploySUS(saLifeCycle, xmlResult);
                }
            }
        }
    }

    /**
     * Undeploy all SUs by calling the "undeploySU" method of the corresponding
     * SUDeploymentManager.
     * 
     * @param saLifeCycle
     *            the parent {@link ServiceAssemblyLifeCycle}
     * @param xmlResult
     *            xml task result
     * @throws Exception
     *             if an error occurs during the undeployment of a service unit
     */
    protected void undeploySUS(ServiceAssemblyLifeCycle saLifeCycle, XMLResult xmlResult)
            throws JBIException {
        /*
         * Undeploy the SUs only if the SA is in SHUTDOWN or UNKNOW state
         */
        String saState = saLifeCycle.getCurrentState();
        if (saState != null
                && (LifeCycleMBean.SHUTDOWN.equals(saState) || LifeCycleMBean.UNKNOWN
                        .equals(saState))) {
            /*
             * JBI spec specifies that service units must be undeployed in the
             * reverse order than the order of the initial declared list
             */
            List<ServiceUnit> suList = saLifeCycle.getServiceAssembly().getServiceUnit();
            if (suList != null) {
                Collections.reverse(suList);
                StringBuffer componentOutput = new StringBuffer();
                for (ServiceUnit suDesc : suList) {
                    if (suDesc != null) {
                        String xmlReturn = undeploySU(saLifeCycle, suDesc);
                        componentOutput.append(xmlReturn);
                    }
                }
                xmlResult.addComponentTaskResult(componentOutput);
            }
        } else {
            throw new JBIException("Service Assembly can't be undeployed in this state: " + saState);
        }
    }

    /**
     * Process the undeployment of the related service unit.
     * 
     * @param saLifeCycle
     *            the service assembly life cycle
     * @param serviceUnit
     *            the service unit production element
     * @return a {@link String} xml result of the service unit undeployment (see
     *         section 6.9 of the JSR 208)
     * @throws Exception
     *             if an error occurs during the undeployment process : can't
     *             retrieve the {@link ComponentLifeCycle}, can't retrieve the
     *             component {@link ServiceUnitManager} (SUM) or an exception is
     *             thrown during the call of the undeploy method of the SUM
     */
    protected String undeploySU(ServiceAssemblyLifeCycle saLifeCycle, ServiceUnit serviceUnit) {
        String xmlReturn = null;
        String suName = DeploymentUtils.getServiceUnitName(serviceUnit);
        if (suName != null && saLifeCycle.getServiceUnitsLifeCycles() != null) {
            ServiceUnitLifeCycle serviceUnitLifeCycle = saLifeCycle.getServiceUnitsLifeCycles()
                    .get(suName);
            String targetedComponent = serviceUnit.getTarget().getComponentName();
            // Retrieve the component life cycle
            ComponentLifeCycle clc = managerService.getComponentLifeCycleByName(targetedComponent);
            if (clc == null || clc.getComponent() == null) {
                log.error("The component '" + targetedComponent
                        + "' is not deployed. The service unit cannot be undeployed from it.");
            } else {
                // Retrieve the service unit manager of the component
                ServiceUnitManager sum = clc.getComponent().getServiceUnitManager();
                if (sum == null) {
                    log.error("The service unit manager is null for the component '"
                            + targetedComponent
                            + "'. The service units can not be undeployed from it.");
                } else {
                    // Set the context class loader to the component's one just
                    // for the
                    // method invocation
                    try {
                        Thread.currentThread().setContextClassLoader(
                                sum.getClass().getClassLoader());
                        xmlReturn = sum.undeploy(suName, serviceUnitLifeCycle
                                .getServiceUnitRootPath());
                    } catch (DeploymentException e) {
                        log.error("The service unit manager of the component '" + targetedComponent
                                + "' cannot  undeploy a service unit.");
                    } finally {
                        Thread.currentThread().setContextClassLoader(
                                this.getClass().getClassLoader());
                    }
                }
            }
            // Remove the service unit from the service assembly
            saLifeCycle.getServiceUnitsLifeCycles().remove(suName);
        }
        return xmlReturn;
    }

    /**
     * Check if a component task result is valid against an xml schema
     * 
     * @param componentTask
     *            xml of the component task
     * @param xsd
     *            an {@link InputStream} of the xsd file
     * @return true if valid, false if not valid
     */
    protected boolean isValidComponentTask(String componentTask, InputStream xsd) {
        return XMLResult.isValidComponentTask(componentTask, xsd);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.processor.Task#undo(java.util.HashMap)
     */
    public void undo(HashMap<String, Object> context) throws Exception {
        // TODO : redeploy all the service units
    }

}
