/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.management.task.installation.install.component;

import java.util.HashMap;

import javax.jbi.JBIException;
import javax.management.ObjectName;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.api.control.IllegalContentException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.util.Fractal;
import org.ow2.petals.container.ContainerService;
import org.ow2.petals.container.lifecycle.Installer;
import org.ow2.petals.container.lifecycle.InstallerMBean;
import org.ow2.petals.jbi.component.context.ComponentContextImpl;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.ow2.petals.jbi.management.MBeanNames;
import org.ow2.petals.jbi.management.ManagementException;
import org.ow2.petals.jbi.management.admin.AdminServiceMBean;
import org.ow2.petals.jbi.management.task.AbstractLoggableTask;
import org.ow2.petals.jbi.management.task.TaskContextConstants;
import org.ow2.petals.jbi.management.task.installation.InstallationContextConstants;
import org.ow2.petals.kernel.api.server.PetalsException;
import org.ow2.petals.kernel.server.FractalHelper;
import org.ow2.petals.util.LoggingUtil;

/**
 * This task creates the InstallerMBean for a component and register it in the
 * JMX server <br/>
 * The task has got a non-robust approach : it stops at the first subtask failed<br />
 * and have the undo capability.<br />
 * Inputs : {@link Jbi}, component install root URI,
 * {@link ComponentContextImpl} <br/>
 * Outputs : component installer {@link ObjectName} <br/>
 * 
 * @author ofabre
 * @author Frederic Gardes - eBM Websourcing
 */
public class CreateAndRegisterInstallerMBeanTask extends AbstractLoggableTask {
    /**
     * JBI Container :: JMX Agent Admin
     */
    private AdminServiceMBean adminServiceMBean;

    protected ContainerService containerService;

    /**
     * Creates a new instance of {@link CreateAndRegisterInstallerMBeanTask}
     * 
     * @param log
     * @param adminServiceMBean
     * @param logger
     * @param router
     * @param containerService
     */
    public CreateAndRegisterInstallerMBeanTask(LoggingUtil log,
            AdminServiceMBean adminServiceMBean, ContainerService container) {
        super(log);
        this.adminServiceMBean = adminServiceMBean;
        this.containerService = container;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.processor.Task#execute(java.util.HashMap)
     */
    public void execute(HashMap<String, Object> context) throws Exception {
        ObjectName result = createAndRegisterInstallerMBean(context);
        context.put(InstallationContextConstants.INSTALLER_OBJECT_NAME, result);
    }

    /**
     * Create and register a new InstallerMBean for the loaded component
     * 
     * @param componentContext
     *            the new Component execution context
     * @param descriptor
     *            the component Jbi
     * @param installationRoot
     *            the URI representing the component installation folder
     * @return the new InstallerMBean name
     * @throws PetalsException
     */
    private ObjectName createAndRegisterInstallerMBean(HashMap<String, Object> ctxt)
            throws PetalsException {
        ObjectName result = null;
        try {
            Jbi descriptor = (Jbi) ctxt.get(TaskContextConstants.DESCRIPTOR);
            org.ow2.petals.jbi.descriptor.original.generated.Component componentDescription = descriptor
                    .getComponent();
            MBeanNames mbeanNames = adminServiceMBean.getMBeanNames();
            result = mbeanNames.createInstallerMBeanName(componentDescription.getIdentification()
                    .getName());
            // Create the Fractal installer
            Component installer = containerService.createInstaller(componentDescription);
            // Start the Fractal installer
            FractalHelper.startComponent(installer);
            // Initialize the Fractal installer
            ((InstallerMBean) installer.getFcInterface("service")).init(ctxt);
            // Register the InstallerMBean
            adminServiceMBean.registerInstaller((Installer) installer.getFcInterface("/content"));
        } catch (ManagementException e) {
            throw new PetalsException("Error creating or registring installerMBean", e);
        } catch (NoSuchInterfaceException e) {
            throw new PetalsException("Error creating or registring installerMBean", e);
        } catch (IllegalLifeCycleException e) {
            throw new PetalsException("Error creating or registring installerMBean", e);
        } catch (JBIException e) {
            throw new PetalsException("Error creating or registring installerMBean", e);
        }
        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.processor.Task#undo(java.util.HashMap)
     */
    public void undo(HashMap<String, Object> context) throws Exception {
        ContentController containerContentController = null;
        org.ow2.petals.jbi.descriptor.original.generated.Component componentDescription = null;
        if (context.containsKey(InstallationContextConstants.INSTALLER_OBJECT_NAME)) {
            ObjectName installerMBeanName = (ObjectName) context
                    .get(InstallationContextConstants.INSTALLER_OBJECT_NAME);
            if (installerMBeanName != null) {
                if (adminServiceMBean != null) {
                    try {
                        // Unregister the InstallerMBean
                        adminServiceMBean.unregisterInstaller(installerMBeanName);
                    } catch (ManagementException e) {
                        log.error("Component installer can't be unregistered or doesn't exist", e);
                    }
                }
                if (containerService != null) {
                    // Get the component controller
                    Component container = containerService.getParentContainer();
                    if (container != null) {
                        try {
                            containerContentController = Fractal.getContentController(container);
                            // Get the lifecycle component
                            Jbi descriptor = (Jbi) context.get(TaskContextConstants.DESCRIPTOR);
                            componentDescription = descriptor.getComponent();
                            if (containerContentController != null && componentDescription != null
                                    && componentDescription.getIdentification() != null) {
                                Component installer = FractalHelper.getRecursiveComponentByName(
                                        containerContentController, "Installer_"
                                                + componentDescription.getIdentification()
                                                        .getName());
                                if (installer != null) {
                                    // Stop the Fractal installer
                                    try {
                                        FractalHelper.stopComponent(installer);
                                    } catch (NoSuchInterfaceException e) {
                                        log
                                                .error(
                                                        "Component installer can't be unregistered or doesn't exist",
                                                        e);
                                    } catch (IllegalLifeCycleException e) {
                                        log
                                                .error(
                                                        "Component installer can't be unregistered or doesn't exist",
                                                        e);
                                    }
                                    // Remove the Fractal installer
                                    try {
                                        containerContentController.removeFcSubComponent(installer);
                                    } catch (IllegalContentException e) {
                                        log
                                                .error(
                                                        "Component installer can't be unregistered or doesn't exist",
                                                        e);
                                    } catch (IllegalLifeCycleException e) {
                                        log
                                                .error(
                                                        "Component installer can't be unregistered or doesn't exist",
                                                        e);
                                    }
                                }
                            }
                        } catch (NoSuchInterfaceException e) {
                            log.error("Component installer can't be unregistered or doesn't exist",
                                    e);
                        }
                    }
                }
            }
        }
    }
}
