/**
 * PETALS - PETALS Services Platform. Copyright (c) 2006 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id: PackageHandler.java 154 2 mai 2006 ofabre $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.management.util;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.zip.ZipFile;

import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.jbi.descriptor.JBIDescriptorException;
import org.ow2.petals.jbi.descriptor.original.JBIDescriptorBuilder;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.ow2.petals.kernel.api.server.PetalsException;
import org.ow2.petals.platform.repository.RepositoryService;
import org.ow2.petals.platform.repository.RepositoryServiceImpl.EntityType;
import org.ow2.petals.util.LoggingUtil;
import org.ow2.petals.util.ZipUtil;

import static org.ow2.petals.jbi.descriptor.original.JBIDescriptorBuilder.JBI_DESCRIPTOR_RESOURCE;

/**
 * Utility class for handling packages and package URIs
 * 
 * @author Olivier Fabre - EBM Websourcing
 * @author Roland Naudin - EBM Websourcing
 */
public class PackageHelper {

    private LoggingUtil log;

    private RepositoryService repositorySrv;

    public PackageHelper(Logger logger, RepositoryService repositorySrv) {
        super();
        this.log = new LoggingUtil(logger);
        this.repositorySrv = repositorySrv;
    }

    /**
     * Load the JBI descriptor from the given archive
     * 
     * @param archiveURI
     *            the archive URI
     * @return the {@link Jbi} object
     * @throws PetalsException
     */
    public Jbi loadDescriptor(URI archiveURI) throws PetalsException {
        log.start();

        Jbi result = null;
        File tmpJBIFile = null;

        try {
            /*
             * open the package Zip archive file and extract jbi.xml temp file
             */
            try {
                ZipFile zipArchive = ZipUtil.openZipFile(archiveURI);
                tmpJBIFile = ZipUtil.getEntryAsTemp(zipArchive, JBI_DESCRIPTOR_RESOURCE);
                zipArchive.close();
            } catch (IOException e) {
                throw new PetalsException("Error during jbi.xml extraction from package : "
                        + archiveURI.getPath(), e);
            }

            /*
             * use builder to parse and validate jbi.xml against the schema
             */
            result = JBIDescriptorBuilder.buildJavaJBIDescriptor(tmpJBIFile.toURI().toURL()
                    .openStream());

        } catch (JBIDescriptorException e) {
            throw new PetalsException(e);
        } catch (MalformedURLException e) {
            // can not happen
        } catch (IOException e) {
            throw new PetalsException("Can not open the jbi.xml file in the package: "
                    + archiveURI.getPath());
        } finally {
            // delete temporary file
            if (tmpJBIFile != null) {
                tmpJBIFile.delete();
            }
        }

        log.end();
        return result;
    }

    /**
     * Load the JBI descriptor from the repository folder of the given entity
     * (component, shared libraries or service assembly).
     * 
     * @param entityName
     *            name of entity
     * @param entityType
     *            type of entity (component, shared libraries or service
     *            assembly)
     * @return the {@link Jbi} object
     * @throws PetalsException
     */
    public Jbi loadDescriptor(String entityName, EntityType entityType) throws PetalsException {
        log.start();

        Jbi result = null;

        /*
         * extract descriptor file to parse and validate
         */
        File tmjJBIFile = null;

        try {
            /*
             * Retrieve jbi.xml file from repository
             */
            switch (entityType) {
                case COMPONENT_TYPE:
                    tmjJBIFile = new File(repositorySrv.getComponentInstallRoot(entityName),
                            JBI_DESCRIPTOR_RESOURCE);
                    break;
                case SA_TYPE:
                    tmjJBIFile = new File(repositorySrv.getServiceAssemblyInstallRoot(entityName),
                            JBI_DESCRIPTOR_RESOURCE);
                    break;
                case SL_TYPE:
                    tmjJBIFile = new File(repositorySrv.getSharedLibInstallRoot(entityName),
                            JBI_DESCRIPTOR_RESOURCE);
                    break;
            }

            if (!tmjJBIFile.exists()) {
                throw new PetalsException(
                        "jbi.xml descriptor entry was not found in repository for JBI artifact: "
                                + entityName);
            }

            /*
             * use builder to parse and validate jbi.xml against the schema
             */
            result = JBIDescriptorBuilder.buildJavaJBIDescriptor(tmjJBIFile.toURI().toURL()
                    .openStream());

        } catch (JBIDescriptorException e) {
            throw new PetalsException(e);
        } catch (MalformedURLException e) {
            // can not happen
        } catch (IOException e) {
            throw new PetalsException(
                    "Can not open the jbi.xml file in the repository for the JBI artifact: "
                            + entityName);
        }

        log.end();
        return result;
    }

    /**
     * Load the jbi descriptor from the unzipped package folder of the given
     * entity (component, shared lib or service assembly).
     * 
     * @param unzipRoot
     *            unzipped package root in work directory
     * @return the {@link Jbi} object
     * @throws PetalsException
     */
    public Jbi loadDescriptor(File unzipRoot) throws PetalsException {
        log.start();

        Jbi result = null;

        /*
         * Extract descriptor file to parse and validate
         */
        File tmpJBIFile = new File(unzipRoot, JBI_DESCRIPTOR_RESOURCE);

        try {
            if (!tmpJBIFile.exists()) {
                throw new PetalsException("jbi.xml descriptor not found in the package: "
                        + unzipRoot.getAbsolutePath());
            }

            /*
             * Use builder to parse and validate jbi.xml against the schema
             */
            result = JBIDescriptorBuilder.buildJavaJBIDescriptor(tmpJBIFile.toURI().toURL()
                    .openStream());

        } catch (JBIDescriptorException e) {
            throw new PetalsException(e);
        } catch (MalformedURLException e) {
            // can not happen
        } catch (IOException e) {
            throw new PetalsException("Can not open the jbi.xml file in the package: "
                    + unzipRoot.getAbsolutePath());
        }

        log.end();
        return result;
    }

    /**
     * Process the given ZIP archive in URL string form and transform it into
     * URI form.
     * 
     * @param archiveZipURL
     *            The ZIP archive
     * @param checkIfFileExist
     *            Check if the URL locates a existing
     * @return <code>URI</code> the package URI.
     */
    public URI getPackageURI(String archiveZipURL, boolean checkIfFileExist) throws PetalsException {
        log.start(archiveZipURL);

        URI uri = null;

        /*
         * Create URI
         */
        // Replace all space by %20
        archiveZipURL = archiveZipURL.replaceAll(" ", "%20");
        try {
            URL url = new URL(archiveZipURL);
            uri = url.toURI();
        } catch (MalformedURLException mue) {
            throw new PetalsException("Wrong package archive location specified. URL: "
                    + archiveZipURL, mue);
        } catch (URISyntaxException use) {
            throw new PetalsException("Syntax error converting to URI from given location. URL: "
                    + archiveZipURL, use);
        }

        /*
         * check if the file exists
         */
        if (checkIfFileExist) {
            File archive = new File(uri);
            if (!archive.exists()) {
                throw new PetalsException("JBI package file not found. URL: " + archiveZipURL);
            }
        }

        log.end();
        return uri;
    }

}
