/**
 * PETALS - PETALS Services Platform. Copyright (c) 2008 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.messaging.routing.module;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.LinkedList;
import java.util.List;

import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.util.LoggingUtil;

/**
 * Load the modules and instanciate from a configuration file.
 * 
 * @author Christophe HAMERLING - eBM WebSourcing
 * @since 2.2
 * 
 */
public class RouterModuleLoader {

    public static final String CONFIGURATION_FILE = "router.cfg";

    private final List<RouterModule> modules;

    private LoggingUtil logger;

    /**
     * Creates a new instance of RouterModuleLoader
     * 
     * @param routerService
     */
    public RouterModuleLoader(Logger logger) {
        this.logger = new LoggingUtil(logger);
        this.modules = new LinkedList<RouterModule>();
    }

    /**
     * Load the router modules in the order they have been defined in the
     * configuration
     */
    public void load() {
        final URL authURL = this.getClass()
                .getResource("/" + RouterModuleLoader.CONFIGURATION_FILE);
        if (authURL != null) {
            final List<String> modulesClasses = this.getModuleClasses(authURL);

            // instantiate the listed router modules
            for (final String className : modulesClasses) {
                try {
                    RouterModule module = null;
                    Class<?> clazz = Class.forName(className);
                    if (clazz != null) {
                        final Object o = clazz.newInstance();
                        if (o instanceof RouterModule) {
                            module = (RouterModule) o;
                            this.modules.add(module);
                        }
                    }
                } catch (ClassNotFoundException e) {
                    this.logger.warning("Failed to find the Router module class '" + className
                            + "'", e);
                } catch (InstantiationException e) {
                    this.logger.warning("Failed to instantiate the Router module class '"
                            + className + "'", e);
                } catch (IllegalAccessException e) {
                    this.logger.warning("Can not access to the Router module class '" + className
                            + "'", e);
                }
            }
        }
    }

    /**
     * 
     * @param configurationURL
     * @return
     */
    private List<String> getModuleClasses(URL configurationURL) {
        final List<String> result = new LinkedList<String>();

        try {
            final BufferedReader br = new BufferedReader(new FileReader(new File(configurationURL
                    .toURI())));
            String line = null;
            while ((line = br.readLine()) != null) {
                // skip lines with '#'
                if ((line != null) && (line.length() > 0) && (line.trim().charAt(0) != '#')) {
                    this.logger.debug("Add router module to list '" + line + "'");
                    result.add(line.trim());
                }
            }
        } catch (final FileNotFoundException e) {
            this.logger.warning(e.getMessage());
        } catch (final IOException e) {
            this.logger.warning(e.getMessage());
        } catch (final URISyntaxException e) {
            this.logger.warning(e.getMessage());
        }
        return result;
    }

    /**
     * @return the modules
     */
    public List<RouterModule> getModules() {
        return this.modules;
    }
}
