/**
 * PETALS - PETALS Services Platform. Copyright (c) 2008 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
package org.ow2.petals.platform.persistence;

import java.io.File;
import java.io.IOException;
import java.util.Collection;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Map;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;

import net.sf.joafip.entity.EnumFilePersistenceCloseAction;
import net.sf.joafip.java.util.PHashMap;
import net.sf.joafip.java.util.PHashSet;
import net.sf.joafip.service.DataAccessSession;
import net.sf.joafip.service.FilePersistence;
import net.sf.joafip.service.FilePersistenceException;

/**
 * An automatically persisted optimize for a JNDI support. Note that all
 * returned values from the map changed must be reset to the map to be
 * effectively persisted. TODO: fractalize, handle errors, refactor the
 * synchronization mechanism.
 * 
 * @author rnaudin
 * @since Petals 2.3
 * 
 */
public class JNDIPersistentMap<K, V extends PHashMap<String, Object>> implements Map<K, V> {

    /**
     * Serial UID
     */
    private static final long serialVersionUID = -4836122991329325954L;

    private static final String MAP = "map";

    private Set<K> mapNames;

    private DataAccessSession session;

    private Map<K, V> maps;

    private Map<K, V> diffMaps;

    private final FilePersistence filePersistence;

    private AtomicBoolean persistSynchronizer;

    private AtomicInteger workSynchronizer;

    private Object lockerPersistence;

    private AtomicBoolean timerTriggered;

    private Timer timer;

    private class SaveTask extends TimerTask {

        @SuppressWarnings("unchecked")
        @Override
        public void run() {
            JNDIPersistentMap.this.preparePersistence();
            try {
                JNDIPersistentMap.this.session.closeAndWait(EnumFilePersistenceCloseAction.SAVE);
                JNDIPersistentMap.this.session.open();
                // reload all the map names
                JNDIPersistentMap.this.mapNames = (Set<K>) JNDIPersistentMap.this.session
                        .getObject(MAP);
                // reload all the map in memory
                for (K key : JNDIPersistentMap.this.mapNames) {
                    V value = (V) JNDIPersistentMap.this.session.getObject(key.toString());
                    JNDIPersistentMap.this.maps.put(key, value);
                }
            } catch (FilePersistenceException e) {
                e.printStackTrace();
            } finally {
                JNDIPersistentMap.this.timerTriggered.set(false);
                JNDIPersistentMap.this.finalizePersistence();
            }
        }
    }

    @SuppressWarnings("unchecked")
    public JNDIPersistentMap(File persistenceDirectory) throws IOException,
            FilePersistenceException {
        this.persistSynchronizer = new AtomicBoolean(false);
        this.workSynchronizer = new AtomicInteger(0);
        this.lockerPersistence = new Object();
        this.timerTriggered = new AtomicBoolean(false);
        this.timer = new Timer(true);
        this.diffMaps = new Hashtable<K, V>();
        this.filePersistence = new FilePersistence(persistenceDirectory, false, false, false);
        this.session = this.filePersistence.createDataAccessSession();
        this.session.open();
        this.maps = new Hashtable<K, V>();
        this.mapNames = (PHashSet<K>) this.session.getObject(MAP);
        if (this.mapNames == null) {
            this.mapNames = new PHashSet<K>();
            this.session.setObject(MAP, this.mapNames);
        } else {
            for (K key : this.mapNames) {
                this.maps.put(key, (V) JNDIPersistentMap.this.session.getObject(key.toString()));
            }
        }
    }

    public void clear() {
        synchronized (this.lockerPersistence) {
            this.workSynchronizer.getAndIncrement();
        }

        try {
            if (!this.persistSynchronizer.get()) {
                synchronized (this.maps) {
                    for (K key : this.mapNames) {
                        try {
                            this.session.removeObject(key.toString());
                        } catch (FilePersistenceException e) {
                            e.printStackTrace();
                        }
                    }
                    this.mapNames.clear();
                    this.maps.clear();
                }
                this.save();
            } else {
                this.diffMaps.clear();
            }
        } finally {
            synchronized (this.lockerPersistence) {
                if (this.workSynchronizer.decrementAndGet() == 0 && this.timerTriggered.get()
                        || this.persistSynchronizer.get()) {
                    this.lockerPersistence.notify();
                }
            }
        }

    }

    public boolean containsKey(Object key) {
        boolean result;

        synchronized (this.lockerPersistence) {
            this.workSynchronizer.getAndIncrement();
        }

        // System.out.println("containKey");

        try {
            if (!this.persistSynchronizer.get()) {
                result = this.maps.containsKey(key);
            } else {
                result = this.diffMaps.containsKey(key);
            }
        } finally {
            synchronized (this.lockerPersistence) {
                if (this.workSynchronizer.decrementAndGet() == 0 && this.timerTriggered.get()
                        || this.persistSynchronizer.get()) {
                    this.lockerPersistence.notify();
                }
            }
        }

        return result;
    }

    public boolean containsValue(Object value) {
        return this.maps.containsValue(value);
    }

    public Set<Entry<K, V>> entrySet() {
        return this.maps.entrySet();
    }

    @SuppressWarnings("unchecked")
    public V get(Object key) {
        V result = null;

        // System.out.println("get " + key);

        synchronized (this.lockerPersistence) {
            this.workSynchronizer.getAndIncrement();
        }

        try {
            if (this.persistSynchronizer.get()) {
                // clone to avoid the modification of the original value, which
                // is persisting
                //TODO: dirty code
                PHashMap<String, Object> clone = new PHashMap<String, Object>();
                clone.putAll(this.diffMaps.get(key));
                result = (V)clone;
            } else {
                result = this.maps.get(key);
            }
        } finally {
            synchronized (this.lockerPersistence) {
                if (this.workSynchronizer.decrementAndGet() == 0
                        && (this.timerTriggered.get() || this.persistSynchronizer.get())) {
                    this.lockerPersistence.notify();
                }
            }
        }

        return result;
    }

    public Set<K> keySet() {
        Set<K> result;

        synchronized (this.lockerPersistence) {
            this.workSynchronizer.getAndIncrement();
        }

        try {
            if (!this.persistSynchronizer.get()) {
                result = this.maps.keySet();
            } else {
                result = this.diffMaps.keySet();
            }
        } finally {
            synchronized (this.lockerPersistence) {
                if (this.workSynchronizer.decrementAndGet() == 0
                        && (this.timerTriggered.get() || this.persistSynchronizer.get())) {
                    this.lockerPersistence.notify();
                }
            }
        }

        return result;
    }

    @SuppressWarnings("unchecked")
    public V put(K key, V value) {
        V returnValue = null;

        // System.out.println("put " + key);

        synchronized (this.lockerPersistence) {
            this.workSynchronizer.getAndIncrement();
        }

        try {
            if (this.persistSynchronizer.get()) {
                returnValue = this.diffMaps.put(key, value);
            } else {
                synchronized (this.maps) {
                    returnValue = this.maps.get(key);
                    // reset the value only if it is not the same instance
                    if (returnValue == null || returnValue != value) {
                        try {
                            returnValue = (V) this.session.setObject(key.toString(), value);
                        } catch (FilePersistenceException e) {
                            e.printStackTrace();
                        }
                        this.maps.put(key, value);
                        this.mapNames.add(key);
                    }
                }
                this.save();
            }
        } finally {
            synchronized (this.lockerPersistence) {
                if (this.workSynchronizer.decrementAndGet() == 0
                        && (this.timerTriggered.get() || this.persistSynchronizer.get())) {
                    this.lockerPersistence.notify();
                }
            }
        }

        return returnValue;
    }

    @SuppressWarnings("unchecked")
    public V remove(Object key) {
        V result = null;

        // System.out.println("remove " + key);

        synchronized (this.lockerPersistence) {
            this.workSynchronizer.getAndIncrement();
        }

        try {
            if (this.persistSynchronizer.get()) {
                result = this.diffMaps.remove(key);
            } else {
                synchronized (this.maps) {
                    try {
                        result = (V) this.session.removeObject(key.toString());
                    } catch (FilePersistenceException e) {
                        e.printStackTrace();
                    }
                    this.mapNames.remove(key);
                    this.maps.remove(key);
                }
                this.save();
            }
        } finally {
            synchronized (this.lockerPersistence) {
                if (this.workSynchronizer.decrementAndGet() == 0
                        && (this.timerTriggered.get() || this.persistSynchronizer.get())) {
                    this.lockerPersistence.notify();
                }
            }
        }

        return result;
    }

    public int size() {
        int result;

        synchronized (this.lockerPersistence) {
            this.workSynchronizer.getAndIncrement();
        }

        try {
            if (this.persistSynchronizer.get()) {
                result = this.diffMaps.size();
            } else {
                result = this.maps.size();
            }
        } finally {
            synchronized (this.lockerPersistence) {
                if (this.workSynchronizer.decrementAndGet() == 0
                        && (this.timerTriggered.get() || this.persistSynchronizer.get())) {
                    this.lockerPersistence.notify();
                }
            }
        }

        return result;
    }

    public Collection<V> values() {
        return this.maps.values();
    }

    @Override
    public boolean equals(Object o) {
        return this.maps.equals(o);
    }

    @Override
    public int hashCode() {
        return this.maps.hashCode();
    }

    public boolean isEmpty() {
        boolean result;

        synchronized (this.lockerPersistence) {
            this.workSynchronizer.getAndIncrement();
        }

        try {
            if (this.persistSynchronizer.get()) {
                result = this.diffMaps.isEmpty();
            } else {
                result = this.maps.isEmpty();
            }
        } finally {
            synchronized (this.lockerPersistence) {
                if (this.workSynchronizer.decrementAndGet() == 0
                        && (this.timerTriggered.get() || this.persistSynchronizer.get())) {
                    this.lockerPersistence.notify();
                }
            }
        }

        return result;
    }

    public void putAll(Map<? extends K, ? extends V> putMap) {

        synchronized (this.lockerPersistence) {
            this.workSynchronizer.getAndIncrement();
        }

        try {
            if (this.persistSynchronizer.get()) {
                for (K key : putMap.keySet()) {
                    this.diffMaps.put(key, putMap.get(key));
                }
            } else {
                synchronized (this.maps) {
                    for (K key : putMap.keySet()) {
                        try {
                            this.session.setObject(key.toString(), putMap.get(key));
                        } catch (FilePersistenceException e) {
                            e.printStackTrace();
                        }
                        this.maps.put(key, putMap.get(key));
                        this.mapNames.add(key);
                    }
                }
                this.save();

            }
        } finally {
            synchronized (this.lockerPersistence) {
                if (this.workSynchronizer.decrementAndGet() == 0
                        && (this.timerTriggered.get() || this.persistSynchronizer.get())) {
                    this.lockerPersistence.notify();
                }
            }
        }

        this.save();
    }

    @Override
    public String toString() {
        return this.maps.toString();
    }

    public void shutDown() {
        try {
            // wait that all updates are persisted

            synchronized (lockerPersistence) {
                while (this.workSynchronizer.get() > 0 || this.persistSynchronizer.get()
                        || this.timerTriggered.get()) {
                    System.out.println("workSynchronizer>" + this.workSynchronizer.get());
                    System.out.println("persistSynchronizer>" + this.persistSynchronizer.get());
                    System.out.println("this.timerTriggered>" + this.timerTriggered.get());
                    try {
                        this.lockerPersistence.wait();
                    } catch (InterruptedException e) {
                        e.printStackTrace();
                    }
                }
            }

            this.timer.cancel();
            this.session.close(EnumFilePersistenceCloseAction.DO_NOT_SAVE);
            this.filePersistence.close();
        } catch (FilePersistenceException e) {
            e.printStackTrace();
        }
    }

    /**
     * Critical section where we synchronize the persisted and the diff maps.
     */
    private void preparePersistence() {

        System.out.println("PreparePersistence start");

        synchronized (this.lockerPersistence) {
            while (this.workSynchronizer.get() > 0) {
                try {
                    this.lockerPersistence.wait();
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }

            this.diffMaps.putAll(this.maps);

            this.persistSynchronizer.set(true);
        }

        System.out.println("PreparePersistence end");
    }

    /**
     * Critical section where we synchronize the diff and the persisted maps.
     */
    private void finalizePersistence() {
        System.out.println("finalizePersistence start");

        synchronized (this.lockerPersistence) {
            while (this.workSynchronizer.get() > 0) {
                try {
                    this.lockerPersistence.wait();
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }

            boolean save = false;

            try {
                Set<K> removeKeys = new HashSet<K>(this.mapNames);
                for (K key : this.diffMaps.keySet()) {
                    V value = this.diffMaps.get(key);
                    V oldValue = this.maps.get(key);
                    if (oldValue == null) {
                        // System.out.println(">>>>>>>>Add value");
                        try {
                            this.session.setObject(key.toString(), value);
                        } catch (FilePersistenceException e) {
                            e.printStackTrace();
                        }
                        this.maps.put(key, value);
                        this.mapNames.add(key);
                        save = true;
                        // to be more efficient, the comparison must be done for
                        // each element of the value, because it is a map. Only
                        // the diff must set in the persisted map
                    } else if (!value.equals(oldValue)) {
                        // System.out.println(">>>>>>>>Change value");
                        // TODO: is the remove of the old value is necessary?
                        try {
                            this.session.setObject(key.toString(), value);
                        } catch (FilePersistenceException e) {
                            e.printStackTrace();
                        }
                        this.maps.put(key, value);
                        save = true;
                    } else {
                        // System.out.println(">>>>>>>>Do nothing");
                    }
                    removeKeys.remove(key);
                }

                for (K key : removeKeys) {
                    // System.out.println(">>>>>>>>remove value");
                    try {
                        this.session.removeObject(key.toString());
                    } catch (FilePersistenceException e) {
                        e.printStackTrace();
                    }
                    this.maps.remove(key);
                    this.mapNames.remove(key);
                    save = true;
                }

                this.diffMaps.clear();
                if (save) {
                    this.save();
                }
            } finally {
                this.persistSynchronizer.set(false);
                // notify in case of shutdown
                this.lockerPersistence.notify();
            }
        }
        System.out.println("finalizePersistence end");
    }

    /**
     * Trigger the storing of persistent in one second. It avoid many storing
     * triggering at each modification of the JNDIMap.
     */
    private void save() {
        if (this.timerTriggered.compareAndSet(false, true)) {
            // System.out.println(">>>>>>>>>save");
            this.timer.schedule(new SaveTask(), 1000);
        }
    }
}
