
package org.ow2.petals.util;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * This class is used to format various logs. it can be used as a static class
 * or as a wrapper object of a logger. The methods add to the resulting message
 * a "ClassName-MethodName" information. (only for DEBUG and INFO level)
 * 
 * @author alouis
 * 
 */
public class LoggingUtil {

    protected Logger log;

    protected String name;

    /**
     * Creates a new instance of LoggingUtil
     * 
     * @param logger
     */
    public LoggingUtil(Logger logger) {
        this(logger, "");
    }

    /**
     * Creates a new instance of LoggingUtil
     * 
     * @param logger
     * @param loggerName
     */
    public LoggingUtil(Logger logger, String loggerName) {
        this.log = logger;

        if (loggerName != null && loggerName.trim().length() > 0) {
            loggerName = "[" + loggerName + "] ";
        } else {
            loggerName = "";
        }
        this.name = loggerName;
    }

    /**
     * if log is enabled, Log as <i>DEBUG</i> : -CALL -
     * 
     */
    public void call() {
        if (log != null && log.isLoggable(BasicLevel.DEBUG)) {
            log.log(BasicLevel.LEVEL_DEBUG, name + "-CALL-" + classAndMethod());
        }
    }

    /**
     * if log is enabled, Log as <i>DEBUG</i> : -CALL -
     * 
     * @param msg
     */
    public void call(Object msg) {
        if (log != null && log.isLoggable(BasicLevel.DEBUG)) {
            log.log(BasicLevel.LEVEL_DEBUG, name + "-CALL-" + classAndMethod() + " " + msg);
        }
    }

    /**
     * if log is enabled, Log as <i>DEBUG</i> : -START-
     * 
     */
    public void start() {
        if (log != null && log.isLoggable(BasicLevel.DEBUG)) {
            log.log(BasicLevel.LEVEL_DEBUG, name + "-START-" + classAndMethod());
        }
    }

    /**
     * if log is enabled, Log as <i>DEBUG</i> the message : -START-msg
     * 
     * @param msg
     */
    public void start(Object msg) {
        if (log != null && log.isLoggable(BasicLevel.DEBUG)) {
            log.log(BasicLevel.LEVEL_DEBUG, name + "-START-" + classAndMethod() + " " + msg);
        }
    }

    /**
     * if log is enabled, Log as <i>DEBUG</i> : - END -
     * 
     */
    public void end() {
        if (log != null && log.isLoggable(BasicLevel.DEBUG)) {
            log.log(BasicLevel.LEVEL_DEBUG, name + "-END-" + classAndMethod());
        }
    }

    /**
     * if log is enabled, Log as <i>DEBUG</i> the message : - END -msg
     * 
     * @param msg
     */
    public void end(Object msg) {
        if (log != null && log.isLoggable(BasicLevel.DEBUG)) {
            log.log(BasicLevel.LEVEL_DEBUG, name + "-END-" + classAndMethod() + " " + msg);
        }
    }

    /**
     * if log is enabled, Log as <i>DEBUG</i> the message
     * 
     * @param msg
     */
    public void debug(Object message) {
        if (log != null && log.isLoggable(BasicLevel.DEBUG)) {
            log.log(BasicLevel.LEVEL_DEBUG, name + classAndMethod() + " " + message);
        }
    }

    /**
     * if log is enabled, Log as <i>INFO</i> the message
     * 
     * @param msg
     */
    public void info(Object message) {
        if (log != null && log.isLoggable(BasicLevel.INFO)) {
            if (log.isLoggable(BasicLevel.DEBUG)) {
                log.log(BasicLevel.LEVEL_INFO, name + classAndMethod() + " " + message);
            } else {
                log.log(BasicLevel.LEVEL_INFO, name + message);
            }
        }
    }

    /**
     * if log is enabled, Log as <i>INFO</i> the message and error
     * 
     * @param msg
     * @param error
     */
    public void info(Object message, Throwable error) {
        if (log != null && log.isLoggable(BasicLevel.INFO)) {
            if (log.isLoggable(BasicLevel.DEBUG)) {
                log.log(BasicLevel.LEVEL_INFO, name + classAndMethod() + " " + message, error);
            } else {
                log.log(BasicLevel.LEVEL_INFO, name + message, error);
            }

        }
    }

    /**
     * if log is enabled, Log as warning the message
     * 
     * @param msg
     */
    public void warning(Object message) {
        if (log != null && log.isLoggable(BasicLevel.WARN)) {
            log.log(BasicLevel.LEVEL_WARN, name + message);
        }
    }

    /**
     * if log is enabled, Log as warning the exception
     * 
     * @param msg
     * @param throwable
     */
    public void warning(Object message, Throwable throwable) {
        if (log != null && log.isLoggable(BasicLevel.WARN)) {
            log.log(BasicLevel.LEVEL_WARN, name + message, throwable);
        }
    }

    /**
     * if log is enabled, Log as <i>ERROR</i> the message
     * 
     * @param msg
     */
    public void error(Object message) {
        if (log != null && log.isLoggable(BasicLevel.ERROR)) {
            log.log(BasicLevel.LEVEL_ERROR, name + message);
        }
    }

    /**
     * if log is enabled, Log as <i>ERROR</i> the exception
     * 
     * @param msg
     * @param throwable
     */
    public void error(Object message, Throwable throwable) {
        if (log != null && log.isLoggable(BasicLevel.ERROR)) {
            log.log(BasicLevel.LEVEL_ERROR, name + message, throwable);
        }
    }
    
    /**
     * 
     * @return
     */
    public boolean isDebugEnabled() {
        return isLevelEnabled(BasicLevel.DEBUG);
    }
    
    /**
     * 
     * @return
     */
    public boolean isInfoEnabled() {
        return isLevelEnabled(BasicLevel.INFO);
    }
    
    /**
     * 
     * @return
     */
    public boolean isWarnEnabled() {
        return isLevelEnabled(BasicLevel.WARN);
    }
    
    /**
     * 
     * @return
     */
    public boolean isErrorEnabled() {
        return isLevelEnabled(BasicLevel.ERROR);
    }
    
    private boolean isLevelEnabled(int level) {
        return log != null && log.isLoggable(level);
    }

    // //////////////////////////////////////////////////////////
    // Static part
    // //////////////////////////////////////////////////////////

    /**
     * Create an exception and analyse it in order to find the class and method
     * that called the LoggingUtil method.
     * 
     * @return
     */
    private static String classAndMethod() {
        String result = null;

        // throw and Parse an exception to find in the stack
        // the method that called the Loggingutil

        Throwable t = new Throwable();

        StackTraceElement[] ste = t.getStackTrace();

        if (ste != null && ste.length > 2) {
            StackTraceElement element = ste[2];

            // If the 2nd element in the stack is this class, get the 3rd
            if (element.getClassName().endsWith(LoggingUtil.class.getName())) {
                element = ste[3];
            }
            String className = element.getClassName();

            // remove the package name of the ClassName
            int index = className.lastIndexOf(".");

            if (index > -1) {
                className = className.substring(index + 1, className.length());
            }

            result = className + "." + element.getMethodName() + "()";
        }
        return result;
    }

    public String getName() {
        return this.log.getName();
    }

}
